import base64
import copy
import glob
import json
import os
import re
from decimal import Decimal
from typing import Dict, List, Any, Optional

from cerberus import Validator
from one_table.Utils import cached, get_iso_8601_date


class Table:
    def __init__(self, params: Dict[str, Any]):
        required_params = ["name", "schema", "client"]
        for param in required_params:
            if param not in params:
                raise ValueError(f'Missing {param} property')

        self.debug = params.get('debug', False)
        self.cache = params.get('cache', False)
        self.cache_ttl = params.get('cacheTTLSeg', 30)
        self.cache_path = params.get('cachePath', "./")
        self.json_api = params.get('jsonApi', False)
        self.table_name = params['name']
        self.schema = params['schema']
        self.models = params['schema']['models']
        self.client = params['client'].Table(params['name'])

        self.pk_key = self.schema['indexes']['primary']['hash']
        self.sk_key = self.schema['indexes']['primary']['sort']
        self.filter_key = "filter"
        self.scan_index_forward = "ScanIndexForward"
        self.id_key = "idKey"
        self.consistent_read_key = "ConsistentRead"
        self.filter_expressions = {
            "eq": "=",
            "neq": "<>",
            "gt": ">",
            "gte": ">=",
            "lt": "<",
            "lte": "<="
        }

    def _generate_response(self, raw_response: Dict[str, Any], schema: Dict[str, Any], model: str, limit: int,
                           id_key: str) -> Dict[str, Any]:
        if raw_response['Count'] == 0:
            return {
                "data": [],
                "meta": {
                    "nextPage": None,
                    "itemsPerPage": limit,
                    "count": 0,
                }
            }

        if self.debug:
            print(raw_response)

        items = []
        for item in raw_response['Items']:
            if self.json_api:
                items.append({
                    "id": item.get(id_key, ""),
                    "type": model,
                    "attributes": {k: v for k, v in item.items() if k in schema}
                })
            else:
                items.append({k: v for k, v in item.items() if k in schema})

        next_page = None
        if "LastEvaluatedKey" in raw_response:
            next_page = base64.b64encode(json.dumps(raw_response["LastEvaluatedKey"]).encode('utf-8')).decode('utf-8')

        response = {
            "data" if self.json_api else "items": items,
            "meta": {
                "nextPage" if self.json_api else "nextKey": next_page,
                "itemsPerPage" if self.json_api else "count": limit,
                "count": raw_response['Count'],
            }
        }

        return response

    def _build_key_condition_expression(self, query: Dict[str, Any], hash_key: str, sort_key: Optional[str] = None) -> \
    Dict[str, Any]:
        expression_attribute_names = {'#hash': hash_key}
        expression_attribute_values = {':hashValue': query[hash_key]}
        key_condition_expression = '#hash = :hashValue'

        if sort_key and sort_key in query:
            expression_attribute_names['#sort'] = sort_key
            sort_condition = query[sort_key]

            if isinstance(sort_condition, dict):
                condition = next(iter(sort_condition))
                if condition == 'begins_with':
                    key_condition_expression += ' AND begins_with(#sort, :sortValue)'
                    expression_attribute_values[':sortValue'] = sort_condition[condition]
                # Add more conditions here as needed
            else:
                key_condition_expression += ' AND #sort = :sortValue'
                expression_attribute_values[':sortValue'] = sort_condition

        return {
            'ExpressionAttributeNames': expression_attribute_names,
            'ExpressionAttributeValues': expression_attribute_values,
            'KeyConditionExpression': key_condition_expression
        }

    def _build_filter_expression(self, filter_conditions: List[Dict[str, Any]]) -> Dict[str, Any]:
        expression_attribute_names = {}
        expression_attribute_values = {}
        filter_expressions = []

        for idx, filter_condition in enumerate(filter_conditions):
            for attr, condition in filter_condition.items():
                attr_name = f'#filter{idx}'
                attr_value = f':filter{idx}'
                expression_attribute_names[attr_name] = attr

                if isinstance(condition, dict):
                    operator = next(iter(condition))
                    value = condition[operator]
                    expression_attribute_values[attr_value] = value
                    filter_expressions.append(f"{attr_name} {self.filter_expressions[operator]} {attr_value}")
                else:
                    expression_attribute_values[attr_value] = condition
                    filter_expressions.append(f"{attr_name} = {attr_value}")

        return {
            'ExpressionAttributeNames': expression_attribute_names,
            'ExpressionAttributeValues': expression_attribute_values,
            'FilterExpression': ' AND '.join(filter_expressions)
        }

    def get_model(self, model: str) -> Dict[str, Any]:
        return self.models[model]

    def create(self, model: str, data: Dict[str, Any]) -> Dict[str, Any]:
        _schema = copy.deepcopy(self.models[model])
        sk = self.models[model][self.sk_key]['value']
        pk = self.models[model][self.pk_key]['value']
        del _schema[self.pk_key]
        del _schema[self.sk_key]
        v = Validator(_schema)
        if v.validate(data):
            data[self.pk_key] = pk.format(**data)
            data[self.sk_key] = sk.format(**data)
            data['createdAt'] = get_iso_8601_date()
            data['updatedAt'] = get_iso_8601_date()
            data['_type'] = model
            self.client.put_item(Item=data)
            return {k: v for k, v in data.items() if k in _schema.keys()}
        else:
            raise ValueError(v.errors)

    def create_if_not_exist(self, model: str, data: Dict[str, Any]) -> Dict[str, Any]:
        _schema = copy.deepcopy(self.models[model])
        sk = self.models[model][self.sk_key]['value']
        pk = self.models[model][self.pk_key]['value']
        del _schema[self.pk_key]
        del _schema[self.sk_key]

        v = Validator(_schema)
        if v.validate(data):
            pk_value = pk.format(**data)
            sk_value = sk.format(**data)

            data[self.pk_key] = pk_value
            data[self.sk_key] = sk_value
            data['createdAt'] = get_iso_8601_date()
            data['updatedAt'] = get_iso_8601_date()
            data['_type'] = model

            condition_expression = f'attribute_not_exists({self.pk_key}) AND attribute_not_exists({self.sk_key})'

            try:
                response = self.client.put_item(
                    Item=data,
                    ConditionExpression=condition_expression,
                    ReturnValues='ALL_OLD'
                )

                if self.debug:
                    print(f"Create response: {response}")

                if 'Attributes' in response:
                    # This means the item already existed
                    raise ValueError("Item already exists")

                return {k: v for k, v in data.items() if k in _schema.keys()}

            except self.client.meta.client.exceptions.ConditionalCheckFailedException:
                raise ValueError("Item already exists")

        else:
            raise ValueError(v.errors)

    def update(self, model: str, data: Dict[str, Any]) -> Dict[str, Any]:
        _schema = copy.deepcopy(self.models[model])
        sk = self.models[model][self.sk_key]['value']
        pk = self.models[model][self.pk_key]['value']
        del _schema[self.pk_key]
        del _schema[self.sk_key]
        pk_keys = re.findall(r'\{([A-Za-z0-9_]+)\}', pk)
        sk_keys = re.findall(r'\{([A-Za-z0-9_]+)\}', sk)
        index_keys = pk_keys + sk_keys
        for item in _schema:
            if item in index_keys:
                _schema[item]['required'] = True
            else:
                _schema[item]['required'] = False

        v = Validator(_schema)
        if v.validate(data):
            expression_attribute_names = {}
            expression_attribute_values = {}
            update_expression = []
            data['updated_at'] = get_iso_8601_date()
            for idx, (key, value) in enumerate(data.items()):
                expression_attribute_names[f'#k{idx}'] = key
                expression_attribute_values[f':v{idx}'] = value
                update_expression.append(f'#k{idx} = :v{idx}')

            pk_key_parsed = pk.format(**data)
            sk_key_parsed = sk.format(**data)

            request = {
                'ConditionExpression': f'(attribute_exists({self.pk_key})) AND (attribute_exists({self.sk_key}))',
                'ExpressionAttributeNames': expression_attribute_names,
                'ExpressionAttributeValues': expression_attribute_values,
                'UpdateExpression': f'SET {", ".join(update_expression)}',
                'ReturnValues': 'ALL_NEW',
                'Key': {
                    'pk': pk_key_parsed,
                    'sk': sk_key_parsed
                }
            }
            if self.debug:
                print(request)
            response = self.client.update_item(**request)
            if 'Attributes' in response:
                return {k: v for k, v in response['Attributes'].items() if k in _schema.keys()}
            return response
        else:
            raise ValueError(v.errors)

    def counter(self, model: str, action: str, query: Dict[str, Any], value: int = 1) -> Dict[str, Any]:
        _schema = copy.deepcopy(self.models[model])
        action = "+" if action == "increment" else "-"
        request = {
            "Key": {
                self.pk_key: query[self.pk_key],
                self.sk_key: query[self.sk_key],
            },
            "UpdateExpression": f"SET #att = #att {action} :val",
            "ExpressionAttributeValues": {
                ':val': Decimal(value)
            },
            'ReturnValues': 'ALL_NEW',
            "ExpressionAttributeNames": {
                "#att": query["key"]
            }
        }
        if self.debug:
            print(request)
        response = self.client.update_item(**request)
        return response.get('Attributes')

    @cached()
    def find(self, model: str, query: Dict[str, Any], limit: int = 100, next_key: Optional[str] = None) -> Dict[
        str, Any]:
        _schema = copy.deepcopy(self.models[model])
        del _schema[self.pk_key]
        del _schema[self.sk_key]

        key_condition = self._build_key_condition_expression(query, self.pk_key, self.sk_key)

        request = {
            **key_condition,
            "ConsistentRead": query.get(self.consistent_read_key, False),
            "ScanIndexForward": query.get(self.scan_index_forward, True),
            "Limit": limit,
        }

        if self.filter_key in query:
            filter_expression = self._build_filter_expression(query[self.filter_key])
            request['FilterExpression'] = filter_expression['FilterExpression']
            request['ExpressionAttributeNames'].update(filter_expression['ExpressionAttributeNames'])
            request['ExpressionAttributeValues'].update(filter_expression['ExpressionAttributeValues'])

        if next_key:
            request['ExclusiveStartKey'] = json.loads(base64.b64decode(next_key))

        if self.debug:
            print(request)

        response = self.client.query(**request)
        id_key = query.get(self.id_key, "id")

        return self._generate_response(response, _schema, model, limit, id_key)

    def find_by_gsi(self, index_name: str, model: str, query: Dict[str, Any], limit: int = 100,
                    next_key: Optional[str] = None) -> Dict[str, Any]:
        if index_name not in self.schema['indexes']:
            raise ValueError(f"Index '{index_name}' not found in schema")

        _schema = copy.deepcopy(self.models[model])
        del _schema[self.pk_key]
        del _schema[self.sk_key]

        gsi_hash_key = self.schema['indexes'][index_name]['hash']
        gsi_sort_key = self.schema['indexes'][index_name]['sort']

        key_condition = self._build_key_condition_expression(query, gsi_hash_key, gsi_sort_key)

        request = {
            'IndexName': index_name,
            **key_condition,
            'ScanIndexForward': query.get(self.scan_index_forward, True),
            'ConsistentRead': query.get(self.consistent_read_key, False),
            'Limit': limit
        }

        if self.filter_key in query:
            filter_expression = self._build_filter_expression(query[self.filter_key])
            request['FilterExpression'] = filter_expression['FilterExpression']
            request['ExpressionAttributeNames'].update(filter_expression['ExpressionAttributeNames'])
            request['ExpressionAttributeValues'].update(filter_expression['ExpressionAttributeValues'])

        if next_key:
            request['ExclusiveStartKey'] = json.loads(base64.b64decode(next_key))

        if self.debug:
            print(request)

        response = self.client.query(**request)
        id_key = query.get(self.id_key, "id")

        return self._generate_response(response, _schema, model, limit, id_key)

    def delete(self, query: Dict[str, Any]):
        self.client.delete_item(
            Key={
                self.pk_key: query[self.pk_key],
                self.sk_key: query[self.sk_key],
            }
        )

    def invalidate_cache(self):
        files = glob.glob(f"{self.cache_path}*.one_cache")
        if self.debug:
            print(files)
        for f in files:
            if self.debug:
                print(f)
            os.remove(f)